@@ -41,8 +41,6 @@
 #define GRL_LOG_DOMAIN_DEFAULT metadata_store_log_domain
 GRL_LOG_DOMAIN_STATIC(metadata_store_log_domain);
 
-#define PLUGIN_ID   METADATA_STORE_PLUGIN_ID
-
 #define SOURCE_ID   "grl-metadata-store"
 #define SOURCE_NAME _("Metadata Store")
 #define SOURCE_DESC _("A plugin for storing extra metadata information")
@@ -120,9 +118,11 @@ enum {
   MEDIA_AUDIO,
   MEDIA_VIDEO,
   MEDIA_IMAGE,
-  MEDIA_BOX
+  MEDIA_CONTAINER
 };
 
+static void grl_metadata_store_source_class_finalize (GObject *object);
+
 static GrlMetadataStoreSource *grl_metadata_store_source_new (void);
 
 static void grl_metadata_store_source_resolve (GrlSource *source,
@@ -150,6 +150,7 @@ gboolean grl_metadata_store_source_plugin_init (GrlRegistry *registry,
                                                 GrlPlugin *plugin,
                                                 GList *configs);
 
+G_DEFINE_TYPE (GrlMetadataStoreSource, grl_metadata_store_source, GRL_TYPE_SOURCE);
 
 /* =================== GrlMetadataStore Plugin  =============== */
 
@@ -174,9 +175,18 @@ grl_metadata_store_source_plugin_init (GrlRegistry *registry,
   return TRUE;
 }
 
-GRL_PLUGIN_REGISTER (grl_metadata_store_source_plugin_init,
-                     NULL,
-                     PLUGIN_ID);
+GRL_PLUGIN_DEFINE (GRL_MAJOR,
+                   GRL_MINOR,
+                   METADATA_STORE_PLUGIN_ID,
+                   "Metadata Store",
+                   "A plugin for storing extra metadata information",
+                   "Igalia S.L.",
+                   VERSION,
+                   "LGPL",
+                   "http://www.igalia.com",
+                   grl_metadata_store_source_plugin_init,
+                   NULL,
+                   NULL);
 
 /* ================== GrlMetadataStore GObject ================ */
 
@@ -194,8 +204,11 @@ grl_metadata_store_source_new (void)
 static void
 grl_metadata_store_source_class_init (GrlMetadataStoreSourceClass * klass)
 {
+  GObjectClass *g_class = G_OBJECT_CLASS (klass);
   GrlSourceClass *source_class = GRL_SOURCE_CLASS (klass);
 
+  g_class->finalize = grl_metadata_store_source_class_finalize;
+
   source_class->supported_keys = grl_metadata_store_source_supported_keys;
   source_class->writable_keys = grl_metadata_store_source_writable_keys;
   source_class->get_caps = grl_metadata_store_source_get_caps;
@@ -207,6 +220,16 @@ grl_metadata_store_source_class_init (GrlMetadataStoreSourceClass * klass)
   g_type_class_add_private (klass, sizeof (GrlMetadataStorePrivate));
 }
 
+static void
+grl_metadata_store_source_class_finalize (GObject *object)
+{
+  GrlMetadataStoreSource *source = GRL_METADATA_STORE_SOURCE (object);
+
+  sqlite3_close (source->priv->db);
+
+  G_OBJECT_CLASS (grl_metadata_store_source_parent_class)->finalize (object);
+}
+
 static void
 grl_metadata_store_source_init (GrlMetadataStoreSource *source)
 {
@@ -259,18 +282,14 @@ grl_metadata_store_source_init (GrlMetadataStoreSource *source)
   // For backwards compatibility, add newer columns if they don't exist
   // in the old database.
   sqlite3_exec (source->priv->db, GRL_SQL_ALTER_TABLE_ADD_FAVOURITE,
-                NULL, NULL, &sql_error);
+                NULL, NULL, NULL);
 
   sqlite3_exec (source->priv->db, GRL_SQL_ALTER_TABLE_ADD_TYPE_ID,
-                NULL, NULL, &sql_error);
+                NULL, NULL, NULL);
 
   GRL_DEBUG ("  OK");
 }
 
-G_DEFINE_TYPE (GrlMetadataStoreSource,
-               grl_metadata_store_source,
-               GRL_TYPE_SOURCE);
-
 /* ======================= Utilities ==================== */
 
 static sqlite3_stmt *
@@ -315,8 +334,16 @@ fill_metadata_from_stmt (GrlMedia *media, GList *keys, sqlite3_stmt *stmt)
       rating = sqlite3_column_double (stmt, STORE_RATING);
       grl_media_set_rating (media, rating, 5.00);
     } else if (key == GRL_METADATA_KEY_LAST_PLAYED) {
+      GDateTime *date;
       last_played = (gchar *) sqlite3_column_text (stmt, STORE_LAST_PLAYED);
-      grl_media_set_last_played (media, last_played);
+      date = grl_date_time_from_iso8601 (last_played);
+      if (date) {
+        grl_media_set_last_played (media, date);
+        g_date_time_unref (date);
+      } else {
+        GRL_WARNING ("Unable to set 'last-played', as '%s' date is invalid",
+                     last_played);
+      }
     } else if (key == GRL_METADATA_KEY_LAST_POSITION) {
       last_position = sqlite3_column_int (stmt, STORE_LAST_POSITION);
       grl_media_set_last_position (media, last_position);
@@ -369,17 +396,17 @@ get_column_name_from_key_id (GrlKeyID key_id)
 static int
 get_media_type (GrlMedia *media)
 {
-  if (GRL_IS_MEDIA_AUDIO (media)) {
+  if (grl_media_is_audio (media)) {
     return MEDIA_AUDIO;
   }
-  if (GRL_IS_MEDIA_VIDEO (media)) {
+  if (grl_media_is_video (media)) {
     return MEDIA_VIDEO;
   }
-  if (GRL_IS_MEDIA_IMAGE (media)) {
+  if (grl_media_is_image (media)) {
     return MEDIA_IMAGE;
   }
-  if (GRL_IS_MEDIA_BOX (media)) {
-    return MEDIA_BOX;
+  if (grl_media_is_container (media)) {
+    return MEDIA_CONTAINER;
   }
 
   return MEDIA;
@@ -395,7 +422,6 @@ bind_and_exec (sqlite3 *db,
 	       GrlMedia *media)
 {
   gint r;
-  const gchar *char_value;
   gint int_value;
   double double_value;
   GList *iter_names, *iter_keys;
@@ -433,8 +459,14 @@ bind_and_exec (sqlite3 *db,
 	int_value = grl_media_get_last_position (media);
 	sqlite3_bind_int (stmt, count, int_value);
       } else if (key == GRL_METADATA_KEY_LAST_PLAYED) {
-	char_value = grl_media_get_last_played (media);
-	sqlite3_bind_text (stmt, count, char_value, -1, SQLITE_STATIC);
+	GDateTime *date;
+	char *date_str;
+	date = grl_media_get_last_played (media);
+	if (date) {
+	  date_str = g_date_time_format (date, "%F %T");
+	  sqlite3_bind_text (stmt, count, date_str, -1, SQLITE_STATIC);
+	  g_free (date_str);
+	}
       } else if (key == GRL_METADATA_KEY_FAVOURITE) {
         int_value = (gint) grl_media_get_favourite (media);
         sqlite3_bind_int (stmt, count, int_value);
@@ -637,8 +669,8 @@ create_media (sqlite3_stmt * stmt, GList *keys)
   case MEDIA_IMAGE:
     media = grl_media_image_new ();
     break;
-  case MEDIA_BOX:
-    media = grl_media_box_new ();
+  case MEDIA_CONTAINER:
+    media = grl_media_container_new ();
     break;
   default:
     media = grl_media_new ();
@@ -720,12 +752,12 @@ grl_metadata_store_source_may_resolve (GrlSource *source,
 
 
   if (media) {
-    if (!(GRL_IS_MEDIA_VIDEO (media) ||
-          GRL_IS_MEDIA_AUDIO (media) ||
+    if (!(grl_media_is_video (media) ||
+          grl_media_is_audio (media) ||
           key_id == GRL_METADATA_KEY_FAVOURITE))
       /* the keys we handle for now only make sense for audio and video,
          with exception of the 'favourite' key, valid as well for pictures
-         and boxes */
+         and containers */
       return FALSE;
 
     if (grl_data_has_key (GRL_DATA (media), GRL_METADATA_KEY_ID))
